/* Object Rexx convenience class for making IBM's Bean Scripting Framework (BSF)
   available in an OO-style as well; Object Rexx programmers should appreciate
   this little class ...
   ;)

   date:        2001-04-22, 2001-05-02 (on the airplane, returning home
                                        from the 12th International Rexx Symposium),
                2001-05-07 (fixed bug in bsf:init, if directly used to instantiate data)
                2001-05-26 (made method "bsf.exit" available at the instance level)
                2001-06-02 (BSF: made method "bsf.sleep" available at the instance level and
                            as "sleep" at the class level;
                            BSF_ARRAY: MAKEARRAY and SUPPLIER now only return entries with
                                       values as does Object Rexx with native arrays)
   version:     1.0.4

   authors:     (c) 2001 Rony G. Flatscher, University of Essen, WU (pronounced: "vey-uh") University Vienna
                         Peter Kalender (October 2000 - February 2001), University of Essen

  *  LICENSE:
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
  *  License as published by the Free Software Foundation; either
  *  version 2.1 of the License, or (at your option) any later version.
  *
  *  This library is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  *  Lesser General Public License for more details.
  *
  *  You should have received a copy of the GNU Lesser General Public
  *  License along with this library; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
  *  The following URL may be available for further information:
  *
  *     <a href="http://www.gnu.org/copyleft/lesser.html">
  *              http://www.gnu.org/copyleft/lesser.html
  *     </a>

*/



/* -------------------------------------------------------------
   routine which turns an array into an argument string as expected
   by the Java side, starting with the indicated element
   ------------------------------------------------------------- */
:: ROUTINE makeArgString
   use arg args

   tmpString    = ""
   do i=1 to args~items
      tmpString = tmpString ', "' || args[i] || '"'
   END
   return tmpString


/* -------------------------------------------------------------
   routine which checks whether data returned from the Java side
   represents a Java object; if so, an Object Rexx proxy is created
   and returned, else the received string value
   ------------------------------------------------------------- */
:: ROUTINE checkResult
   parse arg data

   if data=.bsf~null_string then return .nil

        /* if an object was returned, it has been registered on the Java side;
           therefore reuse, if possible or, create an Object Rexx proxy of it;
           first check whether the name was constructed on the Java side        */
  if pos("@", data)>0 then      -- is this a beanName, created by Java?
  do
     parse var data before '@' after
     if before <> "" then
     do
        if datatype(after, "X") then    -- o.k. second part a hexadecimal value?
        do
           if pos(" ", after)=0 then    -- o.k. no space there?
           do
              -- create an Object Rexx proxy object for the Java object
              dimension = verify(data, "[")-1  -- is it a Java array? with 5 or less dimensions?
              if dimension>0 & dimension<6 then
                 return .bsf_array~new(data, dimension) -- return a BSF_ARRAY proxy

              return .bsf_reference~new(data)   -- return a BSF_REFERENCE prox
           end
        end
     end
  end

  return data           -- assume the string of a primitive type




/* *************************************************************
   Class:       IBM's "Bean Scripting Framework"
   Purpose:     the proxy super-class for the Java classes
   ************************************************************* */
::CLASS BSF PUBLIC

/* -------------------------------------------------------------
   class method allows for creating a proxy class for a Java class
   which is made available to this process by placing it into the
   .local environment
   ------------------------------------------------------------- */
::METHOD import class
  USE ARG classname, java_path

  new_class = .BSF~subclass(classname)  -- create a subclass of .BSFClass

                -- create code for the INIT-method of the new class
  method_txt = 'self~init:super("' || java_path || '", ARG(1,"A"))'
  new_class~define('init', method_txt)  -- define the init-code with the class

  .local~setentry(classname, new_class) -- put new class into the local environment


/* -------------------------------------------------------------
   class method which merely initializes the class attribute "null_string"
   to its default value
   ------------------------------------------------------------- */
::METHOD init           CLASS
  EXPOSE null_string
  null_string=".NIL"

/* -------------------------------------------------------------
   class attribute method stores and allows access to the string
   which is used to indicat "null" for the Java side; upon return
   of values from Java, a Java "null" will be represented via
   Object Rexx' .nil object
   ------------------------------------------------------------- */
::METHOD null_string    ATTRIBUTE CLASS


/* -------------------------------------------------------------
   class method shuts down the Java Virtual Machine (JVM), which started
   this instance of Rexx, this causes Rexx to be shut down as well; the
   Java side waits "waitTime" msecs before terminating the JVM
   ------------------------------------------------------------- */
::METHOD exit class
  use arg exitValue, waitTime

  if \datatype(exitValue, "Whole Number") then return BSF("exit")
  if \datatype(waitTime,  "W"           ) then return BSF("exit", exitValue)
  return checkResult(BSF("exit", exitValue, waitTime))

/* -------------------------------------------------------------
   class method to have Java sleep for us "sleepTime" seconds; resolution
   to a msec is calculated, although the Java side may deviate up to 15/100 secs
   (we are not talking realtime here ;)
   ------------------------------------------------------------- */
::METHOD sleep class
  use arg sleepTime

  return checkResult(BSF("sleep", sleepTime))


/* -------------------------------------------------------------
   method to construct an instance (a "bean"); arguments are
   passed to BSF which uses them to create the bean instance
   from the given Java class
   ------------------------------------------------------------- */
::METHOD init
  USE ARG javaClass, args

        -- create an instance of javaClass and register it with BSF,
        -- leave beanName empty, this way we get a unique name from Java
--  code = 'beanName=BSF( "registerBean", , "' || javaClass || '"' makeArgString(ARG(2, "Array")) ")"

  if args~class = .array then tmpStr=makeArgString(args)
                         else tmpSTr=makeArgString(ARG(2, "Array"))

--  code = 'beanName=BSF( "registerBean", , "' || javaClass || '"' makeArgString(args) ")"
  code = 'beanName=BSF( "registerBean", , "' || javaClass || '"' tmpStr ")"

  interpret code        /* execute the code, calling bsf() with 'registerBean'  */

        -- assign beanName to object, use the Java supplied name
  self~ObjectName=beanName

-- say "---> BSF.init  : bean=["||beanName||"]"  -- 2003-01-05, ---rgf
  return self


/* -------------------------------------------------------------
   destructor method (called by Object Rexx when object gets
   destroyed) to unregister the bean from the BSF registry; this
   then allows the JVM to garbage collect that Java object as well
   ------------------------------------------------------------- */
::METHOD uninit
-- say "---> BSF.uninit: bean=["||self~objectname||"]" -- 2003-01-05, ---rgf
  return BSF("unregisterBean", self~objectname)



/* -------------------------------------------------------------
   unknown method (gets called by Object Rexx whenever an unknown
   message was sent to an instance of this class); forwards the
   message to BSF including all of the received arguments
   ------------------------------------------------------------- */
::METHOD unknown
  USE ARG name, args

  code = 'call BSF "invoke", "' || self~objectname || '", "' || name || '"' makeArgString(args)
  interpret code        -- execute this dynamically created Rexx string

  if var("RESULT") then return checkResult(result)      -- check whether primitive type or Java object
                   else return .nil                     -- nothing came back, i.e. Java returned a true "null"




/* -------------------------------------------------------------
   the following methods implement the individual BSF-calls as
   methods, allowing for invoking them in the OO style with the
   help of messages;

   "registerBean" is not made available explicitly, it is
   implicitly part of the creation of a new proxy object (see
   "INIT" method above)

   "unregisterBean" is not made available explicitly, it is
   implicitly part of the destruction of objects (see the
   "UNINIT" methods above)

   "getStaticValue" is not made available to instances as they
   are able to get (and even set) public (static) fields via
   "getFieldValue" resp. "setFieldValue"
   ------------------------------------------------------------- */
--   instance methods

:: METHOD bsf.addEventListener  /* add event listener to the Java object*/
   use arg eventSetName, filter, text
   return BSF( "addEventListener", self~objectname, eventSetName, filter, text )

:: METHOD bsf.exit              /* exit Rexx                    */
   FORWARD MESSAGE ("EXIT") TO (self~class)

:: METHOD bsf.invoke
   parse arg methodName
   -- create UNKNOWN format (first argument is unknown method name)
   return self~unknown(methodName, arg(2, 'A'))


:: METHOD bsf.getFieldValue
   parse arg fieldName
   return checkResult( bsf("getFieldValue", self~objectname, fieldName) )


:: METHOD bsf.setFieldValue
   parse arg fieldName
   code="res=bsf('setFieldValue', '" || self~objectname || "', '" || fieldName || ,
                 "'" makeArgString(arg(2, 'A')) ")"
   interpret code
   return res


:: METHOD bsf.getPropertyValue
   parse arg propertyName, index
   if arg(2, "Omitted") then index = "" -- no index given, set to empty string
   return checkResult( bsf("getPropertyValue", self~objectname, propertyName, index))


:: METHOD bsf.setPropertyValue
   parse arg propertyName, index
   if arg(2, "Omitted") then index = "" -- no index given, set to empty string
   code="res=bsf('setFieldValue', '" || self~objectname || "', '" || propertyName || "', '" ||,
                 index || "'" makeArgString(arg(3, 'A')) ")"
   interpret code
   return res


--   class  methods
:: METHOD     lookupBean                class
   parse arg beanName
   return checkResult( bsf("lookupBean", beanName) )

:: METHOD     pollEventText             class
   parse arg timeOut
   if \datatype(timeOut, "Whole") then return checkResult( bsf("pollEventText"))
   return checkResult( bsf("pollEventText", timeout))    -- with timeout

:: METHOD     getStaticValue            class
   parse arg className, fieldName
   return checkResult( bsf("getStaticValue", className, fieldName))

:: METHOD     postEventText              class
   parse arg eventText
   return bsf("postEventText", eventText)

   -- set the string which from now on represents the Java "null" on both sides,
   -- the Rexx and the Java side
:: METHOD     setRexxNullString         class
   parse arg newValue
   self~null_string = newValue                  -- set class attribute to the new value
   return bsf("setRexxNullString", newValue)    -- set the Java side to the new value



/* *************************************************************
   Class:       BSF_REFERENCE
   Purpose:     the proxy class for objects returned by the Java side
   ************************************************************* */
::CLASS BSF_REFERENCE subclass BSF

/* -------------------------------------------------------------
   method to construct an instance (a "bean"); as the object was returned
   from the Java side, just remember its name, but do not call BSF's init
   again! This proxy object merely allows for sending messages through it
   to the Java object as well. If it goes out of scope, it will be garbage
   collected by Object Rexx which in turn unregisters the object from the
   Java side, decrementing the reference counter there
   ------------------------------------------------------------- */
::METHOD init
   use arg beanName
   self~ObjectName=beanName

-- say "---> BSF_REFERENCE.init  : bean=["||beanName||"]" -- 2003-01-05, ---rgf





/* *************************************************************
   Class:       BSF_ARRAY
   Purpose:     the proxy class for array objects returned by the Java side
   ************************************************************* */
::CLASS BSF_ARRAY     subclass BSF

/* -------------------------------------------------------------
   method to construct an instance (a "bean"); as the object was returned
   from the Java side, just remember its name, but do not call BSF's init
   again! This proxy object merely allows for sending messages through it
   to the Java object as well. If it goes out of scope, it will be garbage
   collected by Object Rexx which in turn unregisters the object from the
   Java side, decrementing the reference counter there
   ------------------------------------------------------------- */
::METHOD init
   expose items dimension      -- set access to object variable
   use arg beanName, dimension  -- assign "dimension" directly to exposed object variable

-- say"init.beanName:" beanName "dimension:" dimension
   self~ObjectName=beanName
   items=-1                     -- set and indicate, that not calculated as of yet


-- say "---> BSF_ARRAY.init  : bean=["||beanName||"]"  -- 2003-01-05, ---rgf


::METHOD dimension
  expose dimension

  if arg(1, "Omitted") then return dimension    -- return number of dimensions

  if arg(1)=1 then return bsf("arrayLength", self~objectname)

  tmp=arg(1)-1         -- determine size (length) of given dimension in Java

  subscripts=SUBWORD("0, 0, 0, 0, 0", 1, tmp)   -- determine needed subscripts
  subscripts= STRIP(subscripts, 'T', ',')              -- make sure no trailing comma

  code="arr=bsf('arrayAt', '" || self~objectname ||  "'," subscripts ")"
-- say "**debug: dimension to look:" arg(1) "code= [" || code || "]"
  INTERPRET code
-- say "arr=" arr

  return bsf("arrayLength", arr)        -- return size of this dimension


  -- Note: This array-proxy is "1"-based (not 0-based as Java), because Object Rexx arrays
  -- are "1", based, therefore subscript adjustments are necessary, before calling the Java
  -- side
::METHOD at
  tmpString=""
  do i=1 to arg()       -- loop over arguments, decrement indices by "1" to get to 0-based indexing
     tmpString=tmpString (arg(i)-1) ","
  end
  tmpString=STRIP(tmpString, "T", ",")  -- remove trailing comma

  code="res=bsf('arrayAt', '" || self~objectName || "'," tmpString  ")"
-- say "***debug at: code [" || code || "]"
  INTERPRET code

  return checkResult( res )


  -- Note: This array-proxy is "1"-based (not 0-based as Java), because Object Rexx arrays
  -- are "1", based, therefore subscript adjustments are necessary, before calling the Java
  -- side
::METHOD put

-- say "**put: arg()="arg() "arg(1)=["arg(1)"] 2=["arg(2)"] 3=["arg(3)"] 4=["arg(4)"] 5=["arg(5)"] 6=["arg(6)"]"

  tmpString=""
  do i=3 to arg()       -- loop over arguments, decrement indices by "1" to get to 0-based indexing
     tmpString=tmpString (arg(i)-1) ","
  end
  tmpString=STRIP(tmpString, "T", ",")  -- remove trailing comma

  code="res=bsf('arrayPut', '" || self~objectName || "', arg(1), arg(2), " tmpString ")"
-- say "***debug at: code [" || code || "]"
  INTERPRET code

  return checkResult( res )

  -- allow for referring to array elements as if it was an Object Rexx array
::METHOD "[]"
  FORWARD MESSAGE ("AT")        -- synonym for "AT", use its implementation


  -- allow for referring to array elements as if it was an Object Rexx array;
  -- using this syntax variant has the following implication: the valueType and the
  -- value itself must be given as *one* string value; the first word is the valueType
  -- the rest the value!
::METHOD "[]="
  expose dimension
  parse arg type value, .
  -- synonym for "PUT", use its implementation
  if      dimension=1 then tmp="arg(2)"
  else if dimension=2 then tmp="arg(2), arg(3)"
  else if dimension=3 then tmp="arg(2), arg(3), arg(4)"
  else if dimension=4 then tmp="arg(2), arg(3), arg(4), arg(5)"
  else if dimension=5 then tmp="arg(2), arg(3), arg(4), arg(5), arg(6)"
  tmp= "arr=.array~of(type, value," tmp ")"
  INTERPRET tmp

  FORWARD MESSAGE ("PUT") ARGUMENTS (arr)



-- calculates (once) and returns number of individual elements in array
::METHOD items
  expose items                  -- set access to object variable
  if items=-1 then
  do
     items=1
     do i=1 to self~dimension
        items=items*self~dimension(i)
     end
  end
  return items



-- this method creates a supplier method
::METHOD  SUPPLIER
  return self~do_the_work(.true)


-- this method is needed by "DO ... OVER"
::METHOD MAKEARRAY
  return self~do_the_work(.false)


-- this method creates either a single (value) array or both a value and an index array
::METHOD DO_THE_WORK PROTECTED
  expose dimension
  use arg bSupplier

  vala=.array~new       -- create a single dimensional array to store the values

  if bSupplier then
     idxa = .array~new  -- create a single dimensional array to store the indices

  if      dimension=1 then call make_1
  else if dimension=2 then call make_2
  else if dimension=3 then call make_3
  else if dimension=4 then call make_4
  else if dimension=5 then call make_5

  if bSupplier then return .supplier~new(vala, idxa)    -- return a supplier object
  return vala                           -- merely return the value array


  make_1:   -- single dimensional array
    w=length(self~items)
    do i=1 to self~items
       if self[i]=.nil then iterate     -- don't process .nils

       if bSupplier then
          idxa[i] = right(i, w)
       vala[i] = self[i]
    end
    return

  make_2:   -- 2-dimensional array
    w1=length(self~dimension(1))
    w2=length(self~dimension(2))
    k=1
    do i1=1 to self~dimension(1)
       do i2=1 to self~dimension(2)
          if self[i1, i2]=.nil then iterate     -- don't process .nils

          if bSupplier then
             idxa[k] = right(i1, w1)||"," right(i2, w2)
          vala[k] = self[i1, i2]
          k=k+1
       end
    end
    return

  make_3:   -- 3-dimensional array
    w1=length(self~dimension(1))
    w2=length(self~dimension(2))
    w3=length(self~dimension(3))
    k=1
    do i1=1 to self~dimension(1)
       do i2=1 to self~dimension(2)
          do i3=1 to self~dimension(3)
             if self[i1, i2, i3]=.nil then iterate      -- don't process .nils

             if bSupplier then
                idxa[k] = right(i1, w1)||"," right(i2, w2) ||"," right(i3, w3)
             vala[k] = self[i1, i2, i3]
             k=k+1
          end
       end
    end
    return

  make_4:   -- 4-dimensional array
    w1=length(self~dimension(1))
    w2=length(self~dimension(2))
    w3=length(self~dimension(3))
    w4=length(self~dimension(4))
    k=1
    do i1=1 to self~dimension(1)
       do i2=1 to self~dimension(2)
          do i3=1 to self~dimension(3)
             do i4=1 to self~dimension(4)
                if self[i1, i2, i3, i4]=.nil then iterate       -- don't process .nils

                if bSupplier then
                   idxa[k] = right(i1, w1)||"," right(i2, w2) ||"," right(i3, w3) ||,
                             "," right(i4, w4)
                vala[k] = self[i1, i2, i3, i4]
                k=k+1
             end
          end
       end
    end
    return

  make_5:   -- 5-dimensional array
    w1=length(self~dimension(1))
    w2=length(self~dimension(2))
    w3=length(self~dimension(3))
    w4=length(self~dimension(4))
    w5=length(self~dimension(5))
    k=1
    do i1=1 to self~dimension(1)
       do i2=1 to self~dimension(2)
          do i3=1 to self~dimension(3)
             do i4=1 to self~dimension(4)
                do i5=1 to self~dimension(5)
                   if self[i1, i2, i3, i4, i5]=.nil then iterate        -- don't process .nils

                   if bSupplier then
                      idxa[k] = right(i1, w1)||"," right(i2, w2) ||"," right(i3, w3) ||,
                                "," right(i4, w4) ||"," right(i5, w5)
                   vala[k] = self[i1, i2, i3, i4, i5]
                   k=k+1
                end
             end
          end
       end
    end
    return




